from tests import (
    tempLogFile,
    checkLog,
    callTracesSampleLogMessage,
    calltracesTestingInfoLogPrefixRegex,
    calltracesTestingWarningLogPrefixRegex,
    calltracesTestingErrorLogPrefixRegex,
)
from calltraces.linetrace import traceError, traceInfo, traceWarning
from calltraces.classtrace import classtrace
from calltraces import commonTraceSettings as traceSettings


@classtrace
class MyClassTraceTestClass:
    def __init__(self, a: int, b: int) -> None:
        self._a = a
        self._b = b

    def publicBaseFunction(self) -> int:

        self._privateFunction()
        self.__superPrivateFunction()

        traceInfo(callTracesSampleLogMessage)

        return self._a**2 + self._b**3

    def _privateFunction(self) -> None:
        self._a += 1

    def __superPrivateFunction(self) -> None:
        self._b += 1

    def publicWrapperFunction(self) -> None:
        return self.publicBaseFunction()


def test_class_trace(tempLogFile, checkLog) -> None:

    testObj = MyClassTraceTestClass(a=3, b=2)

    testObj.publicWrapperFunction()

    logRegularExpression = (
        calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_classTrace\.MyClassTraceTestClass.publicWrapperFunction Starting Function"
        + "\n"
        + traceSettings.LOG_INDENTATION_CHAR
        + calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_classTrace\.MyClassTraceTestClass.publicBaseFunction Starting Function"
        + "\n"
        + traceSettings.LOG_INDENTATION_CHAR
        + traceSettings.LOG_INDENTATION_CHAR
        + calltracesTestingInfoLogPrefixRegex
        + callTracesSampleLogMessage
        + "\n"
        + traceSettings.LOG_INDENTATION_CHAR
        + calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_classTrace\.MyClassTraceTestClass.publicBaseFunction Finished Function"
        + "\n"
        + calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_classTrace\.MyClassTraceTestClass.publicWrapperFunction Finished Function"
    )

    checkLog(
        regularExpression=logRegularExpression,
        tempLogFileToCheck=tempLogFile,
    )



import os
import re
import tempfile

import pytest
from calltraces import commonTraceSettings as traceSettings


@pytest.fixture
def tempLogFile() -> tempfile.NamedTemporaryFile:
    with tempfile.NamedTemporaryFile(
        suffix=".log", prefix=os.path.basename(__file__)
    ) as tf:
        tf_directory = os.path.dirname(tf.name)
        traceSettings.addOutputFileStream(tf.name)
        yield tf


@pytest.fixture
def checkLog():
    def checkLogFunction(
        regularExpression: str, tempLogFileToCheck: tempfile.NamedTemporaryFile
    ) -> None:
        tempLogFileToCheck.seek(0)

        logOutput = tempLogFileToCheck.read().decode(encoding="utf-8").strip()

        logRegex = re.compile(regularExpression)

        assert logRegex.match(logOutput)

    yield checkLogFunction


callTracesSampleLogMessage = "A sample log message to test calltraces..."
calltracesTestingTimeRegex = r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d,\d\d\d "
calltracesTestingInfoLogPrefixRegex = r"\[INFO:\d\d\] " + calltracesTestingTimeRegex
calltracesTestingWarningLogPrefixRegex = (
    r"\[WARNING:\d\d\] " + calltracesTestingTimeRegex
)
calltracesTestingErrorLogPrefixRegex = r"\[ERROR:\d\d\] " + calltracesTestingTimeRegex



import logging
from datetime import datetime
from .traceSettings import traceSettings

commonTraceSettings = traceSettings()

from .linetrace import traceInfo, traceWarning, traceError


def replaceLoggingFunctions():
    logging.info = traceInfo
    logging.warning = traceWarning
    logging.error = traceError
    logging.exception = traceError

    logging.root.info = traceInfo
    logging.root.warning = traceWarning
    logging.root.error = traceError
    logging.root.exception = traceError



from functools import update_wrapper, wraps

from calltraces import commonTraceSettings
from calltraces.functiontrace import __functionDecorator__
from calltraces.linetrace import isDebuggerAttached


def __classDecorator__(cls: any, debug: bool = False):

    if not commonTraceSettings.enabled:
        return cls

    for funcname in dir(cls):
        # Skip private members
        if funcname.startswith("_"):
            continue

        # Get class attributes and check if it is a function
        func = getattr(cls, funcname)
        if not callable(func):
            continue

        # wrap function with functiontrace decorator
        wrapped = __functionDecorator__(func=func, debug=debug)

        # Attach the new functiontrace function in place of the old function
        setattr(cls, funcname, wrapped)
    return cls


def classtrace(cls: any = None, debugOnly: bool = False):
    """
    Trace all object function calls including input and output
    """

    debug = debugOnly and isDebuggerAttached()

    if cls:
        return __classDecorator__(cls=cls, debug=debug)
    else:
        return lambda cls: __classDecorator__(cls=cls, debug=debug)



from tests import (
    tempLogFile,
    checkLog,
    callTracesSampleLogMessage,
    calltracesTestingInfoLogPrefixRegex,
    calltracesTestingWarningLogPrefixRegex,
    calltracesTestingErrorLogPrefixRegex,
)
from calltraces.linetrace import traceError, traceInfo, traceWarning
from calltraces.functiontrace import functiontrace
from calltraces import commonTraceSettings as traceSettings


@functiontrace
def myBaseFunctionToTrace(a: int, b: int) -> int:
    traceInfo(callTracesSampleLogMessage)
    return a**2 + b**3


@functiontrace
def myWrapperFunctionToTrace(a: int, b: int) -> int:
    return myBaseFunctionToTrace(a=a, b=b)


def test_function_trace(tempLogFile, checkLog) -> None:

    myBaseFunctionToTrace(a=3, b=2)

    logRegularExpression = (
        calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_functionTrace\.myBaseFunctionToTrace Starting Function"
        + "\n"
        + traceSettings.LOG_INDENTATION_CHAR
        + calltracesTestingInfoLogPrefixRegex
        + callTracesSampleLogMessage
        + "\n"
        + calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_functionTrace\.myBaseFunctionToTrace Finished Function"
    )

    checkLog(
        regularExpression=logRegularExpression,
        tempLogFileToCheck=tempLogFile,
    )


def test_wrapper_function_trace(tempLogFile, checkLog) -> None:

    myWrapperFunctionToTrace(a=3, b=2)

    logRegularExpression = (
        calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_functionTrace\.myWrapperFunctionToTrace Starting Function"
        + "\n"
        + traceSettings.LOG_INDENTATION_CHAR
        + calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_functionTrace\.myBaseFunctionToTrace Starting Function"
        + "\n"
        + traceSettings.LOG_INDENTATION_CHAR
        + traceSettings.LOG_INDENTATION_CHAR
        + calltracesTestingInfoLogPrefixRegex
        + callTracesSampleLogMessage
        + "\n"
        + traceSettings.LOG_INDENTATION_CHAR
        + calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_functionTrace\.myBaseFunctionToTrace Finished Function"
        + "\n"
        + calltracesTestingInfoLogPrefixRegex
        + r"tests\.test_functionTrace\.myWrapperFunctionToTrace Finished Function"
    )

    checkLog(
        regularExpression=logRegularExpression,
        tempLogFileToCheck=tempLogFile,
    )



from tests import (
    tempLogFile,
    checkLog,
    callTracesSampleLogMessage,
    calltracesTestingInfoLogPrefixRegex,
    calltracesTestingWarningLogPrefixRegex,
    calltracesTestingErrorLogPrefixRegex,
)
from calltraces.linetrace import traceError, traceInfo, traceWarning


def test_info_line_trace(tempLogFile, checkLog) -> None:

    traceInfo(callTracesSampleLogMessage)

    checkLog(
        regularExpression=calltracesTestingInfoLogPrefixRegex
        + callTracesSampleLogMessage,
        tempLogFileToCheck=tempLogFile,
    )


def test_warning_line_trace(tempLogFile, checkLog) -> None:

    traceWarning(callTracesSampleLogMessage)

    checkLog(
        regularExpression=calltracesTestingWarningLogPrefixRegex
        + callTracesSampleLogMessage,
        tempLogFileToCheck=tempLogFile,
    )


def test_error_line_trace(tempLogFile, checkLog) -> None:

    traceError(callTracesSampleLogMessage)

    checkLog(
        regularExpression=calltracesTestingErrorLogPrefixRegex
        + callTracesSampleLogMessage,
        tempLogFileToCheck=tempLogFile,
    )



# Original: https://stackoverflow.com/a/63644860/9168936

from datetime import datetime
from functools import update_wrapper
import logging

from colorama import Back, Fore, Style
from calltraces import commonTraceSettings
from calltraces.linetrace import isDebuggerAttached, traceInfo


def __functionDecorator__(func: any, debug: bool = False):

    funcFullName = f"{func.__module__}.{func.__qualname__}"

    if (
        not commonTraceSettings.enabled
        or funcFullName in commonTraceSettings.ignoreSymbols
        or func.__name__.startswith("_")
    ):
        return func

    logLevel = logging.DEBUG if debug else logging.INFO

    # we define a wrapper function. This will execute all additional code
    # before and after the "real" function.

    def __traceWrapper__(*args, **kwargs):

        output = None
        try:
            traceInfo(f"{funcFullName} Starting Function", logLevel=logLevel)
            if commonTraceSettings.printArguments:
                traceInfo(f"{funcFullName} args: {args}", logLevel=logLevel)
                traceInfo(f"{funcFullName} kwargs: {kwargs}", logLevel=logLevel)

            # Function call
            commonTraceSettings.traceStackDepth = (
                commonTraceSettings.traceStackDepth + 1
            )
            output = func(*args, **kwargs)
            return output
        finally:
            commonTraceSettings.traceStackDepth = (
                commonTraceSettings.traceStackDepth - 1
            )
            traceInfo(f"{funcFullName} Finished Function", logLevel=logLevel)
            if commonTraceSettings.printArguments:
                traceInfo(f"{funcFullName} args: {args}", logLevel=logLevel)
                traceInfo(f"{funcFullName} kwargs: {kwargs}", logLevel=logLevel)
            if commonTraceSettings.printOutputs:
                traceInfo(f"{funcFullName} output: {output}", logLevel=logLevel)

    # Use "update_wrapper" to keep docstrings and other function metadata
    # intact
    update_wrapper(__traceWrapper__, func)

    commonTraceSettings.ignoreSymbols.append(funcFullName)

    # We can now return the wrapped function
    return __traceWrapper__


def functiontrace(function: any = None, debugOnly: bool = False):
    """
    Trace function calls including input and output
    """

    debug = debugOnly and isDebuggerAttached()

    if function:
        return __functionDecorator__(func=function, debug=debug)
    else:
        return lambda func: __functionDecorator__(func=func, debug=debug)



import logging
import os
import sys
from datetime import datetime
from io import TextIOWrapper

from colorama import Back, Fore, Style
from calltraces import commonTraceSettings


def isDebuggerAttached() -> bool:
    gettrace = getattr(sys, "gettrace", None)

    if gettrace is None:
        return False
    elif gettrace():
        return True
    else:
        return False


def getStackDepthSpace(depth: int = 0):
    if depth <= 0:
        depth = commonTraceSettings.traceStackDepth
    return commonTraceSettings.LOG_INDENTATION_CHAR.join(["" for i in range(depth)])


def formatMessageForStackDepth(msg: str, depth: int = 0):
    return msg.replace(
        "\n",
        "\n"
        + commonTraceSettings.LOG_INDENTATION_CHAR
        + getStackDepthSpace(depth=depth),
    )


def traceInfo(msg, **kwargs) -> None:
    """
    Standard print statement that can be piped to a log stream or sent to a cloud logging solution
    """

    if not commonTraceSettings.enabled:
        return

    logLevel = kwargs.pop("logLevel", logging.INFO)

    kwargs["stackDepthSpace"] = getStackDepthSpace()
    kwargs["stackDepthNum"] = f"{commonTraceSettings.traceStackDepth:02}"

    isDebug = (logLevel <= logging.DEBUG) and isDebuggerAttached()

    if isDebug:
        logLevel = logging.INFO

    for color in commonTraceSettings.all_colors:
        color_selector = color
        if isDebug:
            color_selector += ".debug"
        kwargs[color] = commonTraceSettings.color_constants.get("info", {}).get(
            color_selector, Style.RESET_ALL
        )

    commonTraceSettings.logger.log(
        level=logLevel,
        msg=formatMessageForStackDepth(msg=msg),
        # args=args,
        extra=kwargs,
    )


def traceWarning(msg, **kwargs) -> None:
    """
    Standard warning print statement that can be piped to a log stream or sent to a cloud logging solution
    """

    logLevel = kwargs.pop("logLevel", logging.WARNING)

    kwargs["stackDepthSpace"] = getStackDepthSpace()
    kwargs["stackDepthNum"] = f"{commonTraceSettings.traceStackDepth:02}"

    isDebug = (logLevel <= logging.DEBUG) and isDebuggerAttached()

    if isDebug:
        logLevel = logging.WARNING

    for color in commonTraceSettings.all_colors:
        color_selector = color
        if isDebug:
            color_selector += ".debug"
        kwargs[color] = commonTraceSettings.color_constants.get("info", {}).get(
            color_selector, Style.RESET_ALL
        )

    commonTraceSettings.logger.log(
        level=logLevel,
        msg=formatMessageForStackDepth(msg=msg),
        # args=args,
        extra=kwargs,
    )


def traceError(msg, **kwargs) -> None:
    """
    Standard error print statement that can be piped to a log stream or sent to a cloud logging solution
    """

    logLevel = kwargs.pop("logLevel", logging.ERROR)

    kwargs["stackDepthSpace"] = getStackDepthSpace()
    kwargs["stackDepthNum"] = f"{commonTraceSettings.traceStackDepth:02}"

    isDebug = (logLevel <= logging.DEBUG) and isDebuggerAttached()

    if isDebug:
        logLevel = logging.ERROR

    for color in commonTraceSettings.all_colors:
        color_selector = color
        if isDebug:
            color_selector += ".debug"
        kwargs[color] = commonTraceSettings.color_constants.get("info", {}).get(
            color_selector, Style.RESET_ALL
        )

    if type(msg) != str:
        msg = str(msg)

    sys_exc_info = sys.exc_info()

    is_exception = (
        sys_exc_info is not None
        and type(sys_exc_info) == tuple
        and sys_exc_info[0] is not None
    )

    if is_exception:

        msg += f"""\nDetailed Stack Trace: \n"""

        tb_frame = sys_exc_info[2]
        while tb_frame is not None:

            msg += f"""\n{commonTraceSettings.LOG_INDENTATION_CHAR}{str(tb_frame.tb_frame)}\n"""
            msg += f"""\n{commonTraceSettings.LOG_INDENTATION_CHAR}{commonTraceSettings.LOG_INDENTATION_CHAR}Local Variables:\n"""

            f_locals = tb_frame.tb_frame.f_locals

            for varName in f_locals.keys():
                try:
                    varValue = f_locals[varName]
                    if str(type(varValue)).lower().find("dataframe") >= 0:
                        varValue = f"DataFrame with type {type(varValue)}"
                    else:
                        varValue = str(varValue)
                    if len(varValue) > 1000:
                        varValue = varValue[:999] + "..."
                except Exception as ex:
                    varValue = f"Error getting variable {varName} value. {ex}"
                varValue = formatMessageForStackDepth(msg=varValue, depth=3)
                msg += f"""{commonTraceSettings.LOG_INDENTATION_CHAR}{commonTraceSettings.LOG_INDENTATION_CHAR}[Type {type(varValue)}] {varName} = {varValue}\n"""

            tb_frame = tb_frame.tb_next

    commonTraceSettings.logger.log(
        level=logLevel,
        msg=formatMessageForStackDepth(msg=msg),
        # args=args,
        extra=kwargs,
        exc_info=is_exception,  # Get exception information if available
    )



import logging
import os
import pathlib
import sys
from io import TextIOWrapper
from colorama import Fore, Style
from colorama.ansi import AnsiFore


class PipelineLogFilter(logging.Filter):
    def filter(self, record):
        if not hasattr(record, "stackDepthSpace"):
            record.stackDepthSpace = ""
        if not hasattr(record, "stackDepthNum"):
            record.stackDepthNum = ""
        if not hasattr(record, "color1"):
            record.color1 = ""
        if not hasattr(record, "color2"):
            record.color2 = ""
        if not hasattr(record, "color3"):
            record.color3 = ""
        return True


class traceSettings(object):
    """
    Common settings for tracing the program.

    Example:
    ```
        from calltraces.functiontrace import functiontrace

        @functiontrace
        def myFunc(myParam):
            ...
    ```
    """

    enabled = True
    """
    Global flag for enabling tracing
    """

    LOG_INDENTATION_CHAR = "\t"
    """
    Indentation character to use when showing nested traces
    """

    printArguments = False
    """
    Print the function argument values
    """

    printOutputs = False
    """
    Print the function argument values
    """

    traceStackDepth = 1
    """
    How much to indent a function call. Helpful to measure call-stack depth
    """

    all_colors = ["color1", "color2", "color3"]

    FORMAT_CONSOLE = (
        f"%(stackDepthSpace)s"
        + f"%(color1)s"
        # + f"[%(process)d:%(levelname)s:%(stackDepthNum)s] %(asctime)s"
        + f"[%(levelname)s:%(stackDepthNum)s] %(asctime)s"
        + f"%(color2)s "
        + f"%(message)s"
        + f"%(color3)s"
    )

    FORMAT_FILE = (
        f"%(stackDepthSpace)s"
        # + f"[%(process)d:%(levelname)s:%(stackDepthNum)s] %(asctime)s "
        + f"[%(levelname)s:%(stackDepthNum)s] %(asctime)s "
        + f"%(message)s"
    )

    ignoreSymbols = []

    color_constants = {
        "info": {
            "color1": Fore.GREEN,
            "color2": Style.RESET_ALL,
            "color3": "",
            "color1.debug": Fore.CYAN,
            "color2.debug": " DEBUG:",
            "color3.debug": Style.RESET_ALL,
        },
        "warning": {
            "color1": Fore.YELLOW,
            "color2": "",
            "color3": Style.RESET_ALL,
            "color1.debug": Fore.YELLOW,
            "color2.debug": Fore.CYAN + " DEBUG:",
            "color3.debug": Style.RESET_ALL,
        },
        "error": {
            "color1": Fore.RED,
            "color2": "",
            "color3": Style.RESET_ALL,
        },
    }

    def addOutputFileStream(self, outputFileStream: any) -> None:
        outputFullPathUpper = (
            str(pathlib.Path(outputFileStream).absolute()).strip().upper()
        )

        if outputFullPathUpper in self.outputs:
            return  # Skip adding an existing file to the output stream

        os.makedirs(name=os.path.dirname(outputFileStream), exist_ok=True)

        # create console handler and set level to debug
        fh = logging.FileHandler(filename=outputFileStream, encoding="utf-8")
        fh.setLevel(logging.INFO)

        # create formatter
        formatter = logging.Formatter(
            fmt=self.FORMAT_FILE,
        )

        # add formatter to fh
        fh.setFormatter(formatter)

        # add fh to logger
        self.logger.addHandler(fh)

        self.outputs.append(outputFullPathUpper)

    def __init__(self, *outputStreams):
        """
        Any trace setting initializations
        """
        self.traceStackDepth = 1

        # create console handler and set level to debug
        self.logger = logging.getLogger(name="calltraces")
        self.logger.setLevel(logging.INFO)

        # add logging filter
        self.logger.addFilter(filter=PipelineLogFilter())

        # ch = self.logger.handlers[0]
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)

        # create formatter
        formatter = logging.Formatter(
            fmt=self.FORMAT_CONSOLE,
        )

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        self.logger.addHandler(ch)

        self.outputs = []

        for s in outputStreams:
            self.addOutputFileStream(s)
